package edu.tum.cs.srldb.datadict;

import java.io.Serializable;

import edu.tum.cs.srldb.datadict.domain.BooleanDomain;
import edu.tum.cs.srldb.datadict.domain.DiscardedDomain;
import edu.tum.cs.srldb.datadict.domain.Domain;
import kdl.prox3.dbmgr.DataTypeEnum;

public class DDAttribute implements Cloneable, Serializable {
	private static final long serialVersionUID = 1L;
	protected String name;
	protected Domain domain;
	/**
	 * whether this attribute is scheduled for clustering
	 */
	protected boolean doClustering;
	protected Integer numClusters;
	/**
	 * whether this attribute is actually discarded/unused
	 */
	protected boolean discarded;
	/**
	 * the item that owns this attribute (usually a DDObject)
	 */
	protected DDItem owner;
	
	protected DDAttribute(String name) {
		this.name = name;
		this.domain = null;
		this.doClustering = false;
		this.numClusters = null;
		this.discarded = false;
		this.owner = null;
	}
	
	public DDAttribute(String name, Domain domain) {
		this(name);
		this.domain = domain;
	}
	
	public DDAttribute(String name, Domain domain, boolean doClustering) {
		this(name, domain);
		this.doClustering = doClustering;
	}
	
	/**
	 * @param doClustering whether this attribute should be scheduled for clustering, 
	 * replacing all its values in instances with the respective clustering result
	 */
	public void setClustering(boolean doClustering) {
		this.doClustering = doClustering;
	}
	
	public void setClustering(Integer numClusters) {
		this.doClustering = true;
		this.numClusters = numClusters;
	}
	
	public String getName() {
		return name;
	}
	
	public DataTypeEnum getType() {
		return domain.getType();
	}
	
	public boolean requiresClustering() {
		return doClustering;
	}
	
	public Integer getNumClusters() {
		return numClusters;
	}
	
	public Domain getDomain() {
		return domain;
	}
	
	public boolean isBoolean() {
		return domain instanceof BooleanDomain;
	}
	
	/**
	 * marks this attribute as discarded/unused<br>
	 * 
	 * An attribute may eventually be discarded even though it is defined, because,
	 * for example, it requires clustering and too few instances to actually perform
	 * clustering were found in the database.
	 */
	public void discard() {
		discarded = true;
		domain = DiscardedDomain.getInstance(); // avoid wasting space on domain data
	}
	
	public boolean isDiscarded() {
		return this.discarded;
	}
	
	public DDAttribute clone() {
		try {
			return (DDAttribute)super.clone();
		}
		catch (CloneNotSupportedException e) { return null; }		
	}

	public void setName(String name) {
		this.name = name;
	}
	
	public void setDomain(Domain domain) {
		this.domain = domain;
	}
	
	public void setOwner(DDItem item) throws DDException {
		if(owner == null || item == null)
			owner = item;
		else
			throw new DDException("Error: Cannot add attribute " + this.getName() + " to more than one item; previously added to " + this.owner.getName());
	}
	
	public DDItem getOwner() {
		return owner;
	}
	
	public String toString() {
		return String.format("DDAttribute:%s[domain=%s/size=%d, discarded=%s, clustering=%s/%s]", name, domain.getClass().getSimpleName(), domain.getValues().length, Boolean.toString(discarded), Boolean.toString(doClustering), numClusters == null ? "auto" : numClusters.toString());
	}
}
